define('scribe-plugin-italic-command', [], function() {

    /**
     * This plugin adds a command for italic font style
     * using the EM tag.
     */
    return function() {
        // Browser detections
        //  Opera 8.0+
        var isOpera = (!!window.opr && !!opr.addons) || !!window.opera || navigator.userAgent.indexOf(' OPR/') >= 0,
            // Firefox 1.0+
            isFirefox = typeof InstallTrigger !== 'undefined',
            // Safari 3.0+ "[object HTMLElementConstructor]"
            isSafari = /constructor/i.test(window.HTMLElement) || (function(p) {
                return p.toString() === "[object SafariRemoteNotification]";
            })(!window['safari'] || (typeof safari !== 'undefined' && safari.pushNotification)),
            // Internet Explorer 6-11
            isIE = /*@cc_on!@*/ false || !!document.documentMode,
            // Edge 20+
            isEdge = !isIE && !!window.StyleMedia,
            // Chrome 1 - 79
            isChrome = !!window.chrome && (!!window.chrome.webstore || !!window.chrome.runtime),
            // Edge (based on chromium) detection
            isEdgeChromium = isChrome && (navigator.userAgent.indexOf("Edg") != -1),
            // Blink engine detection
            isBlink = (isChrome || isOpera) && !!window.CSS;

        return function(scribe) {
            var nodeName = "EM",
                commandName = "epitalic",
                realCommandName = "italic",
                codeCommand = new scribe.api.SimpleCommand(commandName, nodeName),

                /**
                 * Selects a node including its tags.
                 *
                 * @param {node} node - reference of a text or element node
                 */
                setSelectNode = function(node) {
                    var sel = getSelection(),
                        range = document.createRange();

                    range.selectNode(node);

                    sel.removeAllRanges();
                    sel.addRange(range);
                },

                /**
                 * Returns a valid node.
                 *
                 * @param {node} node - reference of a text or element node
                 */
                getTextNode = function(node) {
                    // nodeType 3 is a text node
                    return (node.nodeType === 3 && !node.firstChild) ? node : node.firstChild;
                },

                /**
                 * Extends selection to whole EM. Returns EM node or undefined
                 *
                 * @param {selection} object - reference to a wrapped scribe editor selection object
                 */
                selectNodeContent = function(selection) {
                    var node,
                        range;

                    // if nothing is selected
                    if (typeof selection.range === 'undefined' || selection.range.collapsed) {
                        node = selection.getContaining(function(testNode) {
                            return (testNode.nodeName === nodeName);
                        });

                        // if a node was found
                        if (node) {
                            range = document.createRange();
                            range.selectNode(node);
                            selection.selection.removeAllRanges();
                            selection.selection.addRange(range);
                        }
                    }

                    return node;
                },

                /**
                 * replace a selection/range with a html fragment
                 *
                 * @param {scribe} object - reference to scribe editor
                 * @param {content} string - html string with the new content
                 * @param {selectContent} boolean - true if content should be selected after insertion
                 */
                replaceSelectionWithHtml = function(scribe, content, selectContent) {
                    var scribeSelection,
                        selection,
                        range,
                        fragment,
                        firstInsertedNode,
                        lastInsertedNode,
                        fragmentRoot,
                        child;

                    // get selection via scribe instance
                    scribeSelection = new scribe.api.Selection();
                    selection = scribeSelection.selection;

                    // check if Selection object contains at least one range
                    if (selection.getRangeAt && selection.rangeCount) {
                        // get the first range and remove it
                        range = selection.getRangeAt(0);
                        range.deleteContents();

                        // check if browser supports range.createContextualFragment method
                        if (range.createContextualFragment) {
                            // create html fragment from content
                            fragment = range.createContextualFragment(content);

                        // workaround for non supporting browsers
                        } else {
                            // create html fragment from content
                            fragmentRoot = document.createElement("div");
                            fragmentRoot.innerHTML = content;
                            fragment = document.createDocumentFragment();

                            // get first child of fragment root
                            child = fragmentRoot.firstChild;

                            while (child) {
                                // append child to fragement (means that child gets removed from fragmentRoot)
                                fragment.appendChild(child);
                                child = fragmentRoot.firstChild;
                            }
                        }

                        // get nodes for selection (=> selectContent === true)
                        firstInsertedNode = fragment.firstChild;
                        lastInsertedNode = fragment.lastChild;

                        // insert new fragment as range
                        range.insertNode(fragment);

                        // select new content
                        if (selectContent) {
                            if (firstInsertedNode) {
                                range.setStartBefore(firstInsertedNode);
                                range.setEndAfter(lastInsertedNode);
                            }

                            // remove already existing ranges
                            selection.removeAllRanges();
                            // add new selection
                            selection.addRange(range);
                        }
                    }
                };

            /**
             * Is called if the italic button was clicked in the editor.
             *
             * @param {value} string - possible content
             */
            codeCommand.execute = function(value) {
                scribe.transactionManager.run(function() {
                    var selection = new scribe.api.Selection(),
                        sel = selection.selection,
                        anchorParent = sel.anchorNode.parentNode,
                        focusParent = sel.focusNode.parentNode,
                        range = selection.range,
                        // iterate over all EM-nodes and test if selection is inside of any EM-node
                        containingNode = Array.prototype.slice.call(scribe.el.querySelectorAll(nodeName)).find(function(item) {
                            return item.contains(anchorParent);
                        }),
                        newSel,
                        newRange;

                    // if selection is inside an EM-node => remove it
                    if (((anchorParent.nodeName.toUpperCase() === nodeName) && (anchorParent === focusParent)) || containingNode) {
                        // ToDo (later):
                        // Split up italic fragment if at least one character is selected
                        // inside a range and the range is completly inside an EM-element.
                        // <em>the content</em> => 'cont' is selectet => <em>the </em>con<em>ent</em>


                        // Safari uses the native `execCommand` to remove italic font style
                        // from EM elements
                        if (isSafari) {
                            // select node content first
                            selectNodeContent(selection);
                            // execute native "italic" command
                            document.execCommand(realCommandName, false, value || null);

                        // other browsers don't support the native `execCommand`
                        // to remove italic font style from EM elements
                        } else {
                            containingNode = containingNode ? containingNode : anchorParent;
                            // select node (with tags included)
                            setSelectNode(containingNode);
                            // insert new html fragment
                            replaceSelectionWithHtml(scribe, containingNode.innerHTML, false);
                        }
                    } else {
                        // nothing selected => nothing to do
                        if (range.toString().length === 0) {
                            return;
                        }

                        // execute native "italic" command
                        document.execCommand(realCommandName, false, value || null);

                        // Firefox needs special handling to select the same range as
                        // on execution start
                        if (isFirefox) {
                            newSel = getSelection();
                            // get next parent node (necessary to select text again in Firefox)
                            anchorParent = newSel.anchorNode.nextSibling ? getTextNode(newSel.anchorNode.nextSibling) : getTextNode(newSel.anchorNode);

                            // re-select range
                            setSelectNode(anchorParent);

                            // TRY THIS IF *setSelectNode* DOESN'T WORK!!!
                            // // remove ranges and add new range to selection
                            // // => button in toolbar gets highlighted
                            // sel.removeAllRanges();
                            // newRange = document.createRange();
                            // newRange.setStart(anchorParent, 0);
                            // newRange.setEnd(anchorParent, anchorParent.length);
                            // getSelection().addRange(newRange);
                        }
                    }
                });
            };

            /**
             * Indicates if selection is inside an EM element.
             *
             * There is no native command for EM elements, so provide
             * an own `queryState` method.
             */
            codeCommand.queryState = function() {
                var selection = new scribe.api.Selection();

                // test if current surrounding node is an EM element
                return !!selection.getContaining(function(node) {
                    return node.nodeName === nodeName;
                }.bind(this));
            };

            /**
             * Indicates if command is available (enabled/disabled) with
             * the current selection.
             *
             * Use native command here.
             */
            codeCommand.queryEnabled = function() {
                // check if selection is inside an italic element
                return document.queryCommandEnabled(realCommandName);
            };

            // register command
            scribe.commands[commandName] = codeCommand;
        };
    };
});